from __future__ import annotations

import asyncio
import inspect
import json
import time
import uuid
import logging
from typing import Any

from pydantic import BaseModel

from .context import ContextView
from .client import NexumClient

logger = logging.getLogger("nexum")


class Worker:
    def __init__(
        self,
        workflows: list,
        *,
        concurrency: int = 4,
        poll_interval: float = 0.1,
        host: str = "localhost",
        port: int = 50051,
    ):
        self._workflows = workflows
        self._workflow_map = {wf.workflow_id: wf for wf in workflows}
        self._version_hashes = {wf.version_hash: wf for wf in workflows}
        self._concurrency = concurrency
        self._poll_interval = poll_interval
        self._client = NexumClient(host=host, port=port)
        self._worker_id = f"py-worker-{uuid.uuid4().hex[:8]}"
        self._running = False
        self._semaphore: asyncio.Semaphore | None = None

    async def run_until_complete(self, execution_id: str, timeout: float = 60) -> dict:
        """Start worker and wait for specific execution to complete."""
        self._running = True
        worker_task = asyncio.create_task(self._run())

        deadline = time.time() + timeout
        try:
            while time.time() < deadline:
                status = self._client.get_status(execution_id)
                if status["status"] in ("COMPLETED", "FAILED", "CANCELLED"):
                    self._running = False
                    worker_task.cancel()
                    if status["status"] == "FAILED":
                        raise RuntimeError(f"Execution {execution_id} failed")
                    return status["completedNodes"]
                await asyncio.sleep(0.5)
        finally:
            self._running = False
            worker_task.cancel()

        raise TimeoutError(f"Execution {execution_id} did not complete within {timeout}s")

    async def _run(self) -> None:
        self._semaphore = asyncio.Semaphore(self._concurrency)

        while self._running:
            try:
                # Poll for each registered workflow version
                for wf in self._workflows:
                    if not self._running:
                        return
                    resp = self._client.poll_task(self._worker_id, wf.version_hash)
                    if resp.has_task:
                        asyncio.ensure_future(self._handle_task(resp, wf))
                        break
                else:
                    await asyncio.sleep(self._poll_interval)
            except Exception as e:
                logger.error(f"Poll error: {e}")
                await asyncio.sleep(1.0)

    async def _handle_task(self, task, wf) -> None:
        async with self._semaphore:
            try:
                await self._execute_task(task, wf)
            except Exception as e:
                logger.error(f"Task {task.task_id} failed: {e}")
                try:
                    self._client.fail_task(task.task_id, str(e))
                except Exception:
                    pass

    async def _execute_task(self, task, wf) -> None:
        node = wf.get_node(task.node_id)
        if node is None:
            raise RuntimeError(f"Unknown node: {task.node_id}")

        # TIMER: auto-completed server-side; if we get here, complete it
        if node.type == "TIMER":
            from datetime import datetime, timezone
            output = {
                "waited_until": datetime.now(timezone.utc).isoformat(),
                "delay_seconds": node.delay_seconds or 0,
            }
            self._client.complete_task(task.task_id, output)
            logger.info(f"[NEXUM] TIMER {node.id} → completed ({node.delay_seconds}s)")
            return

        # Parse input_json — server sends { input: {...}, deps: {...} }
        input_data_raw = json.loads(task.input_json) if task.input_json else {}
        input_data = input_data_raw.get("input", input_data_raw)
        deps_raw = input_data_raw.get("deps", {})

        # Deserialize deps into Pydantic models where possible
        outputs: dict[str, Any] = {}
        for k, v in deps_raw.items():
            dep_node = wf.get_node(k)
            if dep_node and dep_node.output_model and isinstance(v, dict):
                try:
                    outputs[k] = dep_node.output_model.model_validate(v)
                except Exception:
                    outputs[k] = v
            else:
                outputs[k] = v

        ctx = ContextView(input_data=input_data, outputs=outputs)

        logger.info(f"[NEXUM] {node.type} {node.id} → executing")

        if node.handler is None:
            raise RuntimeError(f"Node {node.id} has no handler")

        if inspect.iscoroutinefunction(node.handler):
            result = await node.handler(ctx)
        else:
            result = node.handler(ctx)

        # Validate output with Pydantic
        if node.output_model and not isinstance(result, node.output_model):
            if isinstance(result, dict):
                result = node.output_model.model_validate(result)
            else:
                raise TypeError(f"Handler for {node.id} returned wrong type: {type(result)}")

        # Serialize output
        if isinstance(result, BaseModel):
            output_json = result.model_dump_json()
        else:
            output_json = json.dumps(result)

        self._client.complete_task(task.task_id, json.loads(output_json))
        logger.info(f"[NEXUM] {node.type} {node.id} → completed")


def worker(workflows: list, *, concurrency: int = 4, **kwargs) -> Worker:
    return Worker(workflows, concurrency=concurrency, **kwargs)



from typing import Any


class ContextView:
    def __init__(self, input_data: dict, outputs: dict[str, Any]):
        self.input = input_data
        self._outputs = outputs

    def get(self, node_id: str) -> Any:
        if node_id not in self._outputs:
            raise KeyError(f"Node '{node_id}' not completed yet. Available: {list(self._outputs.keys())}")
        return self._outputs[node_id]

    def get_map_results(self, map_node_id: str) -> list:
        result = self._outputs.get(map_node_id)
        if not isinstance(result, list):
            raise TypeError(f"'{map_node_id}' is not a MAP node or not completed")
        return result



from __future__ import annotations

import json
from typing import Any

import grpc

from .proto import nexum_pb2, nexum_pb2_grpc


class NexumClient:
    def __init__(self, host: str = "localhost", port: int = 50051):
        self._channel = grpc.insecure_channel(f"{host}:{port}")
        self._stub = nexum_pb2_grpc.NexumServiceStub(self._channel)

    def register_workflow(self, wf) -> str:
        """Register a workflow and return compatibility status."""
        req = nexum_pb2.WorkflowIR(
            workflow_id=wf.workflow_id,
            version_hash=wf.version_hash,
            ir_json=wf.ir_json,
        )
        resp = self._stub.RegisterWorkflow(req)
        if not resp.ok:
            raise RuntimeError(f"RegisterWorkflow failed: {resp.message}")
        return resp.compatibility

    def start_execution(self, workflow_id: str, input_data: dict, version_hash: str = "") -> str:
        req = nexum_pb2.StartRequest(
            workflow_id=workflow_id,
            version_hash=version_hash,
            input_json=json.dumps(input_data),
        )
        resp = self._stub.StartExecution(req)
        return resp.execution_id

    def get_status(self, execution_id: str) -> dict:
        req = nexum_pb2.StatusRequest(execution_id=execution_id)
        resp = self._stub.GetStatus(req)
        completed_nodes: dict[str, Any] = {}
        if resp.completed_nodes_json:
            try:
                completed_nodes = json.loads(resp.completed_nodes_json)
            except json.JSONDecodeError:
                pass
        return {
            "status": resp.status,
            "completedNodes": completed_nodes,
        }

    def poll_task(self, worker_id: str, version_hash: str):
        req = nexum_pb2.PollRequest(
            worker_id=worker_id,
            version_hash=version_hash,
        )
        return self._stub.PollTask(req)

    def complete_task(self, task_id: str, output: Any) -> None:
        output_json = json.dumps(output) if not isinstance(output, str) else output
        req = nexum_pb2.CompleteRequest(
            task_id=task_id,
            output_json=output_json,
        )
        self._stub.CompleteTask(req)

    def fail_task(self, task_id: str, error: str) -> None:
        req = nexum_pb2.FailRequest(
            task_id=task_id,
            error_message=error,
        )
        self._stub.FailTask(req)

    def close(self) -> None:
        self._channel.close()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: nexum.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'nexum.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0bnexum.proto\x12\x05nexum\"H\n\nWorkflowIR\x12\x13\n\x0bworkflow_id\x18\x01 \x01(\t\x12\x14\n\x0cversion_hash\x18\x02 \x01(\t\x12\x0f\n\x07ir_json\x18\x03 \x01(\t\"A\n\x0b\x41\x63kResponse\x12\n\n\x02ok\x18\x01 \x01(\x08\x12\x15\n\rcompatibility\x18\x02 \x01(\t\x12\x0f\n\x07message\x18\x03 \x01(\t\"M\n\x0cStartRequest\x12\x13\n\x0bworkflow_id\x18\x01 \x01(\t\x12\x14\n\x0cversion_hash\x18\x02 \x01(\t\x12\x12\n\ninput_json\x18\x03 \x01(\t\"%\n\rStartResponse\x12\x14\n\x0c\x65xecution_id\x18\x01 \x01(\t\"6\n\x0bPollRequest\x12\x11\n\tworker_id\x18\x01 \x01(\t\x12\x14\n\x0cversion_hash\x18\x02 \x01(\t\"\xb8\x02\n\x0cPollResponse\x12\x10\n\x08has_task\x18\x01 \x01(\x08\x12\x0f\n\x07task_id\x18\x02 \x01(\t\x12\x14\n\x0c\x65xecution_id\x18\x03 \x01(\t\x12\x0f\n\x07node_id\x18\x04 \x01(\t\x12\x12\n\ninput_json\x18\x05 \x01(\t\x12\x17\n\x0fidempotency_key\x18\x06 \x01(\t\x12\x11\n\tnode_type\x18\x07 \x01(\t\x12\x15\n\rmap_item_json\x18\n \x01(\t\x12\x16\n\x0eis_map_subtask\x18\x0b \x01(\x08\x12\x11\n\tmap_index\x18\x0c \x01(\x05\x12\x11\n\tmap_total\x18\r \x01(\x05\x12\x18\n\x10sub_execution_id\x18\x0e \x01(\t\x12\x17\n\x0fsub_workflow_id\x18\x0f \x01(\t\x12\x16\n\x0esub_input_json\x18\x10 \x01(\t\"7\n\x0f\x43ompleteRequest\x12\x0f\n\x07task_id\x18\x01 \x01(\t\x12\x13\n\x0boutput_json\x18\x02 \x01(\t\"5\n\x0b\x46\x61ilRequest\x12\x0f\n\x07task_id\x18\x01 \x01(\t\x12\x15\n\rerror_message\x18\x02 \x01(\t\"%\n\rStatusRequest\x12\x14\n\x0c\x65xecution_id\x18\x01 \x01(\t\">\n\x0eStatusResponse\x12\x0e\n\x06status\x18\x01 \x01(\t\x12\x1c\n\x14\x63ompleted_nodes_json\x18\x02 \x01(\t\"A\n\x0bListRequest\x12\x13\n\x0bworkflow_id\x18\x01 \x01(\t\x12\x0e\n\x06status\x18\x02 \x01(\t\x12\r\n\x05limit\x18\x03 \x01(\x05\"w\n\x10\x45xecutionSummary\x12\x14\n\x0c\x65xecution_id\x18\x01 \x01(\t\x12\x13\n\x0bworkflow_id\x18\x02 \x01(\t\x12\x14\n\x0cversion_hash\x18\x03 \x01(\t\x12\x0e\n\x06status\x18\x04 \x01(\t\x12\x12\n\ncreated_at\x18\x05 \x01(\t\";\n\x0cListResponse\x12+\n\nexecutions\x18\x01 \x03(\x0b\x32\x17.nexum.ExecutionSummary\"%\n\rCancelRequest\x12\x14\n\x0c\x65xecution_id\x18\x01 \x01(\t\"*\n\x13ListVersionsRequest\x12\x13\n\x0bworkflow_id\x18\x01 \x01(\t\"\x81\x01\n\x0bVersionInfo\x12\x13\n\x0bworkflow_id\x18\x01 \x01(\t\x12\x14\n\x0cversion_hash\x18\x02 \x01(\t\x12\x15\n\rcompatibility\x18\x03 \x01(\t\x12\x15\n\rregistered_at\x18\x04 \x01(\t\x12\x19\n\x11\x61\x63tive_executions\x18\x05 \x01(\x05\"<\n\x14ListVersionsResponse\x12$\n\x08versions\x18\x01 \x03(\x0b\x32\x12.nexum.VersionInfo\"Z\n\x0e\x41pproveRequest\x12\x14\n\x0c\x65xecution_id\x18\x01 \x01(\t\x12\x0f\n\x07node_id\x18\x02 \x01(\t\x12\x10\n\x08\x61pprover\x18\x03 \x01(\t\x12\x0f\n\x07\x63omment\x18\x04 \x01(\t\"X\n\rRejectRequest\x12\x14\n\x0c\x65xecution_id\x18\x01 \x01(\t\x12\x0f\n\x07node_id\x18\x02 \x01(\t\x12\x10\n\x08\x61pprover\x18\x03 \x01(\t\x12\x0e\n\x06reason\x18\x04 \x01(\t\"\x0e\n\x0c\x45mptyRequest\"e\n\x13PendingApprovalItem\x12\x14\n\x0c\x65xecution_id\x18\x01 \x01(\t\x12\x0f\n\x07node_id\x18\x02 \x01(\t\x12\x13\n\x0bworkflow_id\x18\x03 \x01(\t\x12\x12\n\nstarted_at\x18\x04 \x01(\t\"E\n\x18PendingApprovalsResponse\x12)\n\x05items\x18\x01 \x03(\x0b\x32\x1a.nexum.PendingApprovalItem2\xed\x05\n\x0cNexumService\x12\x39\n\x10RegisterWorkflow\x12\x11.nexum.WorkflowIR\x1a\x12.nexum.AckResponse\x12;\n\x0eStartExecution\x12\x13.nexum.StartRequest\x1a\x14.nexum.StartResponse\x12\x33\n\x08PollTask\x12\x12.nexum.PollRequest\x1a\x13.nexum.PollResponse\x12:\n\x0c\x43ompleteTask\x12\x16.nexum.CompleteRequest\x1a\x12.nexum.AckResponse\x12\x32\n\x08\x46\x61ilTask\x12\x12.nexum.FailRequest\x1a\x12.nexum.AckResponse\x12\x38\n\tGetStatus\x12\x14.nexum.StatusRequest\x1a\x15.nexum.StatusResponse\x12\x39\n\x0eListExecutions\x12\x12.nexum.ListRequest\x1a\x13.nexum.ListResponse\x12;\n\x0f\x43\x61ncelExecution\x12\x14.nexum.CancelRequest\x1a\x12.nexum.AckResponse\x12O\n\x14ListWorkflowVersions\x12\x1a.nexum.ListVersionsRequest\x1a\x1b.nexum.ListVersionsResponse\x12\x38\n\x0b\x41pproveTask\x12\x15.nexum.ApproveRequest\x1a\x12.nexum.AckResponse\x12\x36\n\nRejectTask\x12\x14.nexum.RejectRequest\x1a\x12.nexum.AckResponse\x12K\n\x13GetPendingApprovals\x12\x13.nexum.EmptyRequest\x1a\x1f.nexum.PendingApprovalsResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nexum_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_WORKFLOWIR']._serialized_start=22
  _globals['_WORKFLOWIR']._serialized_end=94
  _globals['_ACKRESPONSE']._serialized_start=96
  _globals['_ACKRESPONSE']._serialized_end=161
  _globals['_STARTREQUEST']._serialized_start=163
  _globals['_STARTREQUEST']._serialized_end=240
  _globals['_STARTRESPONSE']._serialized_start=242
  _globals['_STARTRESPONSE']._serialized_end=279
  _globals['_POLLREQUEST']._serialized_start=281
  _globals['_POLLREQUEST']._serialized_end=335
  _globals['_POLLRESPONSE']._serialized_start=338
  _globals['_POLLRESPONSE']._serialized_end=650
  _globals['_COMPLETEREQUEST']._serialized_start=652
  _globals['_COMPLETEREQUEST']._serialized_end=707
  _globals['_FAILREQUEST']._serialized_start=709
  _globals['_FAILREQUEST']._serialized_end=762
  _globals['_STATUSREQUEST']._serialized_start=764
  _globals['_STATUSREQUEST']._serialized_end=801
  _globals['_STATUSRESPONSE']._serialized_start=803
  _globals['_STATUSRESPONSE']._serialized_end=865
  _globals['_LISTREQUEST']._serialized_start=867
  _globals['_LISTREQUEST']._serialized_end=932
  _globals['_EXECUTIONSUMMARY']._serialized_start=934
  _globals['_EXECUTIONSUMMARY']._serialized_end=1053
  _globals['_LISTRESPONSE']._serialized_start=1055
  _globals['_LISTRESPONSE']._serialized_end=1114
  _globals['_CANCELREQUEST']._serialized_start=1116
  _globals['_CANCELREQUEST']._serialized_end=1153
  _globals['_LISTVERSIONSREQUEST']._serialized_start=1155
  _globals['_LISTVERSIONSREQUEST']._serialized_end=1197
  _globals['_VERSIONINFO']._serialized_start=1200
  _globals['_VERSIONINFO']._serialized_end=1329
  _globals['_LISTVERSIONSRESPONSE']._serialized_start=1331
  _globals['_LISTVERSIONSRESPONSE']._serialized_end=1391
  _globals['_APPROVEREQUEST']._serialized_start=1393
  _globals['_APPROVEREQUEST']._serialized_end=1483
  _globals['_REJECTREQUEST']._serialized_start=1485
  _globals['_REJECTREQUEST']._serialized_end=1573
  _globals['_EMPTYREQUEST']._serialized_start=1575
  _globals['_EMPTYREQUEST']._serialized_end=1589
  _globals['_PENDINGAPPROVALITEM']._serialized_start=1591
  _globals['_PENDINGAPPROVALITEM']._serialized_end=1692
  _globals['_PENDINGAPPROVALSRESPONSE']._serialized_start=1694
  _globals['_PENDINGAPPROVALSRESPONSE']._serialized_end=1763
  _globals['_NEXUMSERVICE']._serialized_start=1766
  _globals['_NEXUMSERVICE']._serialized_end=2515
# @@protoc_insertion_point(module_scope)



"""
Nexum Python SDK Demo
Equivalent of examples/demo/run.ts
"""
import asyncio
import sys
import os
import logging

# Add parent directory to path so we can import the nexum package
sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

from pydantic import BaseModel
from nexum import workflow, NexumClient, worker

logging.basicConfig(level=logging.INFO)


# Schemas
class SearchResult(BaseModel):
    content: str
    score: float


class Summary(BaseModel):
    text: str
    word_count: int


# Handlers
async def do_search(ctx):
    query = ctx.input.get("query", "default")
    await asyncio.sleep(0.1)
    return SearchResult(
        content=f"Found relevant information about: {query}",
        score=0.92,
    )


def do_summarize(ctx):
    result = ctx.get("search")
    text = f"Summary: {result.content} (score: {result.score})"
    return Summary(text=text, word_count=len(text.split()))


# Workflow
research_agent = (
    workflow("ResearchAgentPy")
    .effect("search", SearchResult, handler=do_search)
    .compute("summarize", Summary, handler=do_summarize)
    .build()
)


async def main():
    client = NexumClient()

    # Register workflow
    compat = client.register_workflow(research_agent)
    print(f"Registered workflow: {research_agent.workflow_id} (compatibility: {compat})")

    # Start execution
    exec_id = client.start_execution(
        research_agent.workflow_id,
        {"query": "Durable Execution for LLM Agents"},
        version_hash=research_agent.version_hash,
    )
    print(f"Started execution: {exec_id}")

    # Run worker until execution completes
    w = worker([research_agent], concurrency=2)
    result = await w.run_until_complete(exec_id, timeout=30)

    print(f"\nResult:")
    print(f"  Search: {result.get('search', {})}")
    print(f"  Summary: {result.get('summarize', {})}")
    print("\n[SUCCESS] Python SDK demo completed!")


asyncio.run(main())



from __future__ import annotations

import hashlib
import json
from typing import Any, Callable, Type

from pydantic import BaseModel

from .context import ContextView


class NodeDef:
    def __init__(
        self,
        node_id: str,
        node_type: str,  # COMPUTE, EFFECT, ROUTER, HUMAN_APPROVAL, TIMER
        output_model: Type[BaseModel] | None,
        handler: Callable | None,
        dependencies: list[str],
        delay_seconds: int | None = None,
    ):
        self.id = node_id
        self.type = node_type
        self.output_model = output_model
        self.handler = handler
        self.dependencies = dependencies
        self.delay_seconds = delay_seconds


class WorkflowDef:
    def __init__(self, workflow_id: str, version_hash: str, ir_json: str, nodes: list[NodeDef]):
        self.workflow_id = workflow_id
        self.version_hash = version_hash
        self.ir_json = ir_json
        self.nodes = nodes
        self._node_map = {n.id: n for n in nodes}

    def get_node(self, node_id: str) -> NodeDef | None:
        return self._node_map.get(node_id)


class WorkflowBuilder:
    def __init__(self, workflow_id: str):
        self.workflow_id = workflow_id
        self._nodes: list[NodeDef] = []
        self._node_order: list[str] = []

    def _current_deps(self) -> list[str]:
        """Default dependency = all previous nodes (matches TS SDK behavior)."""
        return list(self._node_order)

    def effect(
        self,
        node_id: str,
        output_model: Type[BaseModel],
        handler: Callable,
        *,
        depends_on: list[str] | None = None,
    ) -> WorkflowBuilder:
        deps = depends_on if depends_on is not None else self._current_deps()
        self._nodes.append(NodeDef(node_id, "EFFECT", output_model, handler, deps))
        self._node_order.append(node_id)
        return self

    def compute(
        self,
        node_id: str,
        output_model: Type[BaseModel],
        handler: Callable,
        *,
        depends_on: list[str] | None = None,
    ) -> WorkflowBuilder:
        deps = depends_on if depends_on is not None else self._current_deps()
        self._nodes.append(NodeDef(node_id, "COMPUTE", output_model, handler, deps))
        self._node_order.append(node_id)
        return self

    def timer(
        self,
        node_id: str,
        delay_seconds: int,
        *,
        depends_on: list[str] | None = None,
    ) -> WorkflowBuilder:
        deps = depends_on if depends_on is not None else self._current_deps()
        self._nodes.append(NodeDef(node_id, "TIMER", None, None, deps, delay_seconds=delay_seconds))
        self._node_order.append(node_id)
        return self

    def build(self) -> WorkflowDef:
        # Build IR JSON matching the TypeScript SDK format
        ir_nodes: dict[str, Any] = {}
        for n in self._nodes:
            node_ir: dict[str, Any] = {
                "type": n.type,
                "dependencies": n.dependencies,
            }
            if n.delay_seconds is not None:
                node_ir["delay_seconds"] = n.delay_seconds
            ir_nodes[n.id] = node_ir
        ir_json = json.dumps({"nodes": ir_nodes})
        version_hash = "sha256:" + hashlib.sha256(ir_json.encode()).hexdigest()
        return WorkflowDef(self.workflow_id, version_hash, ir_json, list(self._nodes))


def workflow(workflow_id: str) -> WorkflowBuilder:
    return WorkflowBuilder(workflow_id)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import nexum_pb2 as nexum__pb2

GRPC_GENERATED_VERSION = '1.78.1'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + ' but the generated code in nexum_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class NexumServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.RegisterWorkflow = channel.unary_unary(
                '/nexum.NexumService/RegisterWorkflow',
                request_serializer=nexum__pb2.WorkflowIR.SerializeToString,
                response_deserializer=nexum__pb2.AckResponse.FromString,
                _registered_method=True)
        self.StartExecution = channel.unary_unary(
                '/nexum.NexumService/StartExecution',
                request_serializer=nexum__pb2.StartRequest.SerializeToString,
                response_deserializer=nexum__pb2.StartResponse.FromString,
                _registered_method=True)
        self.PollTask = channel.unary_unary(
                '/nexum.NexumService/PollTask',
                request_serializer=nexum__pb2.PollRequest.SerializeToString,
                response_deserializer=nexum__pb2.PollResponse.FromString,
                _registered_method=True)
        self.CompleteTask = channel.unary_unary(
                '/nexum.NexumService/CompleteTask',
                request_serializer=nexum__pb2.CompleteRequest.SerializeToString,
                response_deserializer=nexum__pb2.AckResponse.FromString,
                _registered_method=True)
        self.FailTask = channel.unary_unary(
                '/nexum.NexumService/FailTask',
                request_serializer=nexum__pb2.FailRequest.SerializeToString,
                response_deserializer=nexum__pb2.AckResponse.FromString,
                _registered_method=True)
        self.GetStatus = channel.unary_unary(
                '/nexum.NexumService/GetStatus',
                request_serializer=nexum__pb2.StatusRequest.SerializeToString,
                response_deserializer=nexum__pb2.StatusResponse.FromString,
                _registered_method=True)
        self.ListExecutions = channel.unary_unary(
                '/nexum.NexumService/ListExecutions',
                request_serializer=nexum__pb2.ListRequest.SerializeToString,
                response_deserializer=nexum__pb2.ListResponse.FromString,
                _registered_method=True)
        self.CancelExecution = channel.unary_unary(
                '/nexum.NexumService/CancelExecution',
                request_serializer=nexum__pb2.CancelRequest.SerializeToString,
                response_deserializer=nexum__pb2.AckResponse.FromString,
                _registered_method=True)
        self.ListWorkflowVersions = channel.unary_unary(
                '/nexum.NexumService/ListWorkflowVersions',
                request_serializer=nexum__pb2.ListVersionsRequest.SerializeToString,
                response_deserializer=nexum__pb2.ListVersionsResponse.FromString,
                _registered_method=True)
        self.ApproveTask = channel.unary_unary(
                '/nexum.NexumService/ApproveTask',
                request_serializer=nexum__pb2.ApproveRequest.SerializeToString,
                response_deserializer=nexum__pb2.AckResponse.FromString,
                _registered_method=True)
        self.RejectTask = channel.unary_unary(
                '/nexum.NexumService/RejectTask',
                request_serializer=nexum__pb2.RejectRequest.SerializeToString,
                response_deserializer=nexum__pb2.AckResponse.FromString,
                _registered_method=True)
        self.GetPendingApprovals = channel.unary_unary(
                '/nexum.NexumService/GetPendingApprovals',
                request_serializer=nexum__pb2.EmptyRequest.SerializeToString,
                response_deserializer=nexum__pb2.PendingApprovalsResponse.FromString,
                _registered_method=True)


class NexumServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def RegisterWorkflow(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StartExecution(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PollTask(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CompleteTask(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FailTask(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListExecutions(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CancelExecution(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListWorkflowVersions(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ApproveTask(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RejectTask(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetPendingApprovals(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_NexumServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'RegisterWorkflow': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterWorkflow,
                    request_deserializer=nexum__pb2.WorkflowIR.FromString,
                    response_serializer=nexum__pb2.AckResponse.SerializeToString,
            ),
            'StartExecution': grpc.unary_unary_rpc_method_handler(
                    servicer.StartExecution,
                    request_deserializer=nexum__pb2.StartRequest.FromString,
                    response_serializer=nexum__pb2.StartResponse.SerializeToString,
            ),
            'PollTask': grpc.unary_unary_rpc_method_handler(
                    servicer.PollTask,
                    request_deserializer=nexum__pb2.PollRequest.FromString,
                    response_serializer=nexum__pb2.PollResponse.SerializeToString,
            ),
            'CompleteTask': grpc.unary_unary_rpc_method_handler(
                    servicer.CompleteTask,
                    request_deserializer=nexum__pb2.CompleteRequest.FromString,
                    response_serializer=nexum__pb2.AckResponse.SerializeToString,
            ),
            'FailTask': grpc.unary_unary_rpc_method_handler(
                    servicer.FailTask,
                    request_deserializer=nexum__pb2.FailRequest.FromString,
                    response_serializer=nexum__pb2.AckResponse.SerializeToString,
            ),
            'GetStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetStatus,
                    request_deserializer=nexum__pb2.StatusRequest.FromString,
                    response_serializer=nexum__pb2.StatusResponse.SerializeToString,
            ),
            'ListExecutions': grpc.unary_unary_rpc_method_handler(
                    servicer.ListExecutions,
                    request_deserializer=nexum__pb2.ListRequest.FromString,
                    response_serializer=nexum__pb2.ListResponse.SerializeToString,
            ),
            'CancelExecution': grpc.unary_unary_rpc_method_handler(
                    servicer.CancelExecution,
                    request_deserializer=nexum__pb2.CancelRequest.FromString,
                    response_serializer=nexum__pb2.AckResponse.SerializeToString,
            ),
            'ListWorkflowVersions': grpc.unary_unary_rpc_method_handler(
                    servicer.ListWorkflowVersions,
                    request_deserializer=nexum__pb2.ListVersionsRequest.FromString,
                    response_serializer=nexum__pb2.ListVersionsResponse.SerializeToString,
            ),
            'ApproveTask': grpc.unary_unary_rpc_method_handler(
                    servicer.ApproveTask,
                    request_deserializer=nexum__pb2.ApproveRequest.FromString,
                    response_serializer=nexum__pb2.AckResponse.SerializeToString,
            ),
            'RejectTask': grpc.unary_unary_rpc_method_handler(
                    servicer.RejectTask,
                    request_deserializer=nexum__pb2.RejectRequest.FromString,
                    response_serializer=nexum__pb2.AckResponse.SerializeToString,
            ),
            'GetPendingApprovals': grpc.unary_unary_rpc_method_handler(
                    servicer.GetPendingApprovals,
                    request_deserializer=nexum__pb2.EmptyRequest.FromString,
                    response_serializer=nexum__pb2.PendingApprovalsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'nexum.NexumService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('nexum.NexumService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class NexumService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def RegisterWorkflow(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/nexum.NexumService/RegisterWorkflow',
            nexum__pb2.WorkflowIR.SerializeToString,
            nexum__pb2.AckResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def StartExecution(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/nexum.NexumService/StartExecution',
            nexum__pb2.StartRequest.SerializeToString,
            nexum__pb2.StartResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def PollTask(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/nexum.NexumService/PollTask',
            nexum__pb2.PollRequest.SerializeToString,
            nexum__pb2.PollResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CompleteTask(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/nexum.NexumService/CompleteTask',
            nexum__pb2.CompleteRequest.SerializeToString,
            nexum__pb2.AckResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def FailTask(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/nexum.NexumService/FailTask',
            nexum__pb2.FailRequest.SerializeToString,
            nexum__pb2.AckResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/nexum.NexumService/GetStatus',
            nexum__pb2.StatusRequest.SerializeToString,
            nexum__pb2.StatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListExecutions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/nexum.NexumService/ListExecutions',
            nexum__pb2.ListRequest.SerializeToString,
            nexum__pb2.ListResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CancelExecution(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/nexum.NexumService/CancelExecution',
            nexum__pb2.CancelRequest.SerializeToString,
            nexum__pb2.AckResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListWorkflowVersions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/nexum.NexumService/ListWorkflowVersions',
            nexum__pb2.ListVersionsRequest.SerializeToString,
            nexum__pb2.ListVersionsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ApproveTask(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/nexum.NexumService/ApproveTask',
            nexum__pb2.ApproveRequest.SerializeToString,
            nexum__pb2.AckResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RejectTask(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/nexum.NexumService/RejectTask',
            nexum__pb2.RejectRequest.SerializeToString,
            nexum__pb2.AckResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetPendingApprovals(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/nexum.NexumService/GetPendingApprovals',
            nexum__pb2.EmptyRequest.SerializeToString,
            nexum__pb2.PendingApprovalsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)
